"""Test functions, function, function signatures, etc."""

from __future__ import print_function

try:
    import vmtest
except ImportError:
    from . import vmtest

from xdis import PYTHON_VERSION

class TestFunctions(vmtest.VmTestCase):

    def test_no_builtins(self):
        self.self_checking()

    def test_different_globals_may_have_different_builtins(self):
        self.do_one()

    # test_pos_args has function syntax added in 3.3
    if PYTHON_VERSION >= 3.3:
        def test_pos_args(self):
            self.self_checking()

    if PYTHON_VERSION >= 3.6:
        print("Test not gone over yet for >= 3.6")
    else:

        def test_functions(self):
            self.assert_ok("""\
                def fn(a, b=17, c="Hello", d=[]):
                    d.append(99)
                    print(a, b, c, d)
                fn(1)
                fn(2, 3)
                fn(3, c="Bye")
                fn(4, d=["What?"])
                fn(5, "b", "c")
                """)

        def test_recursion(self):
            self.assert_ok("""\
                def fact(n):
                    if n <= 1:
                        return 1
                    else:
                        return n * fact(n-1)
                f6 = fact(6)
                print(f6)
                assert f6 == 720
                """)

        def test_nested_names(self):
            self.assert_ok("""\
                def one():
                    x = 1
                    def two():
                        x = 2
                        print(x)
                    two()
                    print(x)
                one()
                """)

        def test_calling_functions_with_args_kwargs(self):
            self.assert_ok("""\
                def fn(a, b=17, c="Hello", d=[]):
                    d.append(99)
                    print(a, b, c, d)
                fn(6, *[77, 88])
                fn(**{'c': 23, 'a': 7})
                fn(6, *[77], **{'c': 23, 'd': [123]})
                """)

        def test_defining_functions_with_args_kwargs(self):
            self.do_one()

        def test_defining_functions_with_empty_args_kwargs(self):
            self.assert_ok("""\
                def fn(*args):
                    print("args is %r" % (args,))
                fn()
                """)
            self.assert_ok("""\
                def fn(**kwargs):
                    print("kwargs is %r" % (kwargs,))
                fn()
                """)
            self.assert_ok("""\
                def fn(*args, **kwargs):
                    print("args is %r, kwargs is %r" % (args, kwargs))
                fn()
                """)

        def test_partial(self):
            self.assert_ok("""\
                from _functools import partial

                def f(a,b):
                    return a-b

                f7 = partial(f, 7)
                four = f7(3)
                assert four == 4
                """)

        # FIXME for 3.6
        if PYTHON_VERSION < 3.6:
            def test_partial_with_kwargs(self):
                self.self_checking()

        def test_wraps(self):
            self.assert_ok("""\
                from functools import wraps
                def my_decorator(f):
                    dec = wraps(f)
                    def wrapper(*args, **kwds):
                        print('Calling decorated function')
                        return f(*args, **kwds)
                    wrapper = dec(wrapper)
                    return wrapper

                @my_decorator
                def example():
                    '''Docstring'''
                    return 17

                assert example() == 17
                """)




    class TestClosures(vmtest.VmTestCase):
        def test_closures(self):
            self.self_checking()
        if PYTHON_VERSION >= 3.6:
            print("Test not gone over yet for >= 3.6")
        else:
            if PYTHON_VERSION > 3.3:
                # Investigate for < 3.3
                def test_closure_vars_from_static_parent(self):
                    self.self_checking()

            if PYTHON_VERSION > 3.5:
                # Investigate for < 3.5
                def test_call_ex_kw(self):
                    self.self_checking()

            def test_a_closure(self):
                self.assert_ok("""\
                    def make_adder(x):
                        def add(y):
                            return x+y
                        return add
                    a = make_adder(10)
                    print(a(7))
                    assert a(7) == 17
                    """)

            def test_closures_store_deref(self):
                self.assert_ok("""\
                    def make_adder(x):
                        z = x+1
                        def add(y):
                            return x+y+z
                        return add
                    a = make_adder(10)
                    print(a(7))
                    assert a(7) == 28
                    """)

            def test_closures_in_loop(self):
                self.assert_ok("""\
                    def make_fns(x):
                        fns = []
                        for i in range(x):
                            fns.append(lambda i=i: i)
                        return fns
                    fns = make_fns(3)
                    for f in fns:
                        print(f())
                    assert (fns[0](), fns[1](), fns[2]()) == (0, 1, 2)
                    """)

            def test_closures_with_defaults(self):
                self.assert_ok("""\
                    def make_adder(x, y=13, z=43):
                        def add(q, r=11):
                            return x+y+z+q+r
                        return add
                    a = make_adder(10, 17)
                    print(a(7))
                    assert a(7) == 88
                    """)

    class TestGenerators(vmtest.VmTestCase):
        def test_first(self):
            self.assert_ok("""\
                def two():
                    yield 1
                    yield 2
                for i in two():
                    print(i)
                """)

        def test_partial_generator(self):
            self.assert_ok("""\
                from _functools import partial

                def f(a,b):
                    num = a+b
                    while num:
                        yield num
                        num -= 1

                f2 = partial(f, 2)
                three = f2(1)
                assert list(three) == [3,2,1]
                """)

        def test_yield_multiple_values(self):
            self.assert_ok("""\
                def triples():
                    yield 1, 2, 3
                    yield 4, 5, 6

                for a, b, c in triples():
                    print(a, b, c)
                """)

        def test_simple_generator(self):
            self.assert_ok("""\
                g = (x for x in [0,1,2])
                print(list(g))
                """)

        def test_generator_from_generator(self):
            self.assert_ok("""\
                g = (x*x for x in range(5))
                h = (y+1 for y in g)
                print(list(h))
                """)

        def test_generator_from_generator2(self):
            self.assert_ok("""\
                class Thing(object):
                    RESOURCES = ('abc', 'def')
                    def get_abc(self):
                        return "ABC"
                    def get_def(self):
                        return "DEF"
                    def resource_info(self):
                        for name in self.RESOURCES:
                            get_name = 'get_' + name
                            yield name, getattr(self, get_name)

                    def boom(self):
                        #d = list((name, get()) for name, get in self.resource_info())
                        d = [(name, get()) for name, get in self.resource_info()]
                        return d

                print(Thing().boom())
                """)

        if PYTHON_VERSION >= 3.3:
            # yield from starts in 3.3
            def test_yield_from(self):
                self.assert_ok("""\
                    def main():
                        x = outer()
                        next(x)
                        y = x.send("Hello, World")
                        print(y)

                    def outer():
                        yield from inner()

                    def inner():
                        y = yield
                        yield y

                    main()
                    """)

            def test_yield_from_tuple(self):
                self.do_one()

            def test_distinguish_iterators_and_generators(self):
                self.assert_ok("""\
                    class Foo(object):
                        def __iter__(self):
                            return FooIter()

                    class FooIter(object):
                        def __init__(self):
                            self.state = 0

                        def __next__(self):
                            if self.state >= 10:
                                raise StopIteration
                            self.state += 1
                            return self.state

                        def send(self, n):
                            print("sending")

                    def outer():
                        yield from Foo()

                    for x in outer():
                        print(x)
                    """)

            def test_nested_yield_from(self):
                self.assert_ok("""\
                    def main():
                        x = outer()
                        next(x)
                        y = x.send("Hello, World")
                        print(y)

                    def outer():
                        yield from middle()

                    def middle():
                        yield from inner()

                    def inner():
                        y = yield
                        yield y

                    main()
                    """)

            def test_return_from_generator(self):
                self.assert_ok("""\
                    def gen():
                        yield 1
                        return 2

                    x = gen()
                    while True:
                        try:
                            print(next(x))
                        except StopIteration as e:
                            print(e.value)
                            break
                """)

            def test_return_from_generator_with_yield_from(self):
                self.assert_ok("""\
                    def returner():
                        if False:
                            yield
                        return 1

                    def main():
                        y = yield from returner()
                        print(y)

                    list(main())
                """)

if __name__ == "__main__":
    # import unittest
    # unittest.main()

    t = TestFunctions("test_functions")
    t.test_functions()



import sys

author='Rocky Bernstein, Ned Batchelder, Paul Swartz, Allison Kaptur and others'
author_email='rb@dustyfeet.com'
entry_points = {"console_scripts": ["xpython=xpython.__main__:main"]}

# Python-version | package | last-version |
# -----------------------------------------
# 3.2            | click   | 4.0          |
# 3.2            | pip     | 8.1.2        |
# 3.3            | pip     | 10.0.1       |
# 3.4            | pip     | 19.1.1       |

import os.path as osp
def get_srcdir():
    filename = osp.normcase(osp.dirname(osp.abspath(__file__)))
    return osp.realpath(filename)

def read(*rnames):
    return open(osp.join(get_srcdir(), *rnames)).read()

exec(read("xpython/version.py"))

PYTHON_VERSION = sys.version_info[0] + (sys.version_info[1] / 10.0)
IS_PYPY = '__pypy__' in sys.builtin_module_names

supported_versions = (
    SUPPORTED_PYPY if IS_PYPY else SUPPORTED_PYTHON
)
mess = "PYPY 2.7, 3.2, 3.5 and 3.6" if IS_PYPY else "CPython 2.7, 3.2 .. 3.7"

if PYTHON_VERSION not in supported_versions:
    python = "PyPy " if IS_PYPY else "C"
    print("We only support %s; you have %sPython %s" % (mess, python, PYTHON_VERSION))
    raise Exception(mess)

if 3.0 <= PYTHON_VERSION <= 3.2:
    click_version = '<= 4.0'
else:
    click_version = ''

install_requires=[
    "six",
    "xdis >= 4.6.0, < 4.7.0",
    "click%s" % click_version
],

py_modules = None
short_desc = "Python cross-version byte-code interpeter"
url='http://github.com/rocky/xpython'

classifiers = [
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]

# Get/set VERSION and long_description from files
long_description = read("README.rst") + "\n"
exec(read("xpython/version.py"))



"""Test exceptions."""

from __future__ import print_function
import unittest
try:
    import vmtest
except ImportError:
    from . import vmtest

import six

from xdis import PYTHON_VERSION, PYTHON3
PY2 = not PYTHON3


class TestExceptions(vmtest.VmTestCase):

    def test_raise_exception(self):
        self.assert_ok("raise Exception('oops')", raises=Exception)

    def test_raise_exception_class(self):
        self.assert_ok("raise ValueError", raises=ValueError)

    if PYTHON_VERSION >= 3.6:
        print("Test not gone over yet for >= 3.6")
    else:

        def test_catching_exceptions(self):
            # Catch the exception precisely
            self.assert_ok("""\
                try:
                    [][1]
                    print("Shouldn't be here...")
                except IndexError:
                    print("caught it!")
                """)
            # Catch the exception by a parent class
            self.assert_ok("""\
                try:
                    [][1]
                    print("Shouldn't be here...")
                except Exception:
                    print("caught it!")
                """)
            # Catch all exceptions
            self.assert_ok("""\
                try:
                    [][1]
                    print("Shouldn't be here...")
                except:
                    print("caught it!")
                """)

        if PY2:
            def test_raise_exception_2args(self):
                self.assert_ok("raise ValueError, 'bad'", raises=ValueError)

            def test_raise_exception_3args(self):
                self.assert_ok("""\
                    from sys import exc_info
                    try:
                        raise Exception
                    except:
                        _, _, tb = exc_info()
                    raise ValueError, "message", tb
                    """, raises=ValueError)

        def test_raise_and_catch_exception(self):
            self.assert_ok("""\
                try:
                    raise ValueError("oops")
                except ValueError as e:
                    print("Caught: %s" % e)
                print("All done")
                """)

        if PYTHON3:
            def test_raise_exception_from(self):
                self.assert_ok(
                    "raise ValueError from NameError",
                    raises=ValueError
                )

        def test_raise_and_catch_exception_in_function(self):
            self.assert_ok("""\
                def fn():
                    raise ValueError("oops")

                try:
                    fn()
                except ValueError as e:
                    print("Caught: %s" % e)
                print("done")
                """)

        def test_global_name_error(self):
            self.assert_ok("fooey", raises=NameError)
            self.assert_ok("""\
                try:
                    fooey
                    print("Yes fooey?")
                except NameError:
                    print("No fooey")
                """)

            def test_local_name_error(self):
                self.assert_ok("""\
                    def fn():
                        fooey
                    fn()
                    """, raises=NameError)

        def test_catch_local_name_error(self):
            self.assert_ok("""\
                def fn():
                    try:
                        fooey
                        print("Yes fooey?")
                    except NameError:
                        print("No fooey")
                fn()
                """)

        def test_reraise(self):
            self.assert_ok("""\
                def fn():
                    try:
                        fooey
                        print("Yes fooey?")
                    except NameError:
                        print("No fooey")
                        raise
                fn()
                """, raises=NameError)

        def test_reraise_explicit_exception(self):
            self.assert_ok("""\
                def fn():
                    try:
                        raise ValueError("ouch")
                    except ValueError as e:
                        print("Caught %s" % e)
                        raise
                fn()
                """, raises=ValueError)

        def test_finally_while_throwing(self):
            self.assert_ok("""\
                def fn():
                    try:
                        print("About to..")
                        raise ValueError("ouch")
                    finally:
                        print("Finally")
                fn()
                print("Done")
                """, raises=ValueError)

        def test_coverage_issue_92(self):
            self.assert_ok("""\
                l = []
                for i in range(3):
                    try:
                        l.append(i)
                    finally:
                        l.append('f')
                    l.append('e')
                l.append('r')
                print(l)
                assert l == [0, 'f', 'e', 1, 'f', 'e', 2, 'f', 'e', 'r']
                """)

if __name__ == "__main__":
    unittest.main()



# See https://github.com/nedbat/byterun/pull/33:
#
# Because instances of Falsy are falsy, `func.im_self` in
# call_function() of xpython/pyvm2.ptreated them as nonexistent and
# did not add the self parameter to the argument list.
#
# The problem was exhibited on Python 2.7.
class Falsy(object):
    def __bool__(self):
        return False

    __nonzero__ = __bool__

    def do_stuff(self):
        # See the HitchHiker's guide
        return 42


assert Falsy().do_stuff() == 42



"""
Implementation of a part of the standard Python opcodes.

The rest, dealing with variables in optimized ways, is in nestedscope.py.
"""

from xpython.interpreter.baseobjspace import W_Root


CANNOT_CATCH_MSG = ("catching classes that don't inherit from BaseException "
                    "is not allowed in 3.x")

def unaryoperation(operationname):
    def opimpl(self, *ignored):
        operation = getattr(self.space, operationname)
        w_1 = self.popvalue()
        w_result = operation(w_1)
        self.pushvalue(w_result)
    opimpl.unaryop = operationname

    return func_with_new_name(opimpl, "opcode_impl_for_%s" % operationname)

def binaryoperation(operationname):
    def opimpl(self, *ignored):
        operation = getattr(self.space, operationname)
        w_2 = self.popvalue()
        w_1 = self.popvalue()
        w_result = operation(w_1, w_2)
        self.pushvalue(w_result)
    opimpl.binop = operationname

    return func_with_new_name(opimpl, "opcode_impl_for_%s" % operationname)


class ExitFrame(Exception):
    pass


class Return(ExitFrame):
    """Raised when exiting a frame via a 'return' statement."""


class Yield(ExitFrame):
    """Raised when exiting a frame via a 'yield' statement."""


class RaiseWithExplicitTraceback(Exception):
    """Raised at interp-level by a 0-argument 'raise' statement."""
    def __init__(self, operr):
        self.operr = operr


### Frame Blocks ###

class SuspendedUnroller(W_Root):
    """Abstract base class for interpreter-level objects that
    instruct the interpreter to change the control flow and the
    block stack.

    The concrete subclasses correspond to the various values WHY_XXX
    values of the why_code enumeration in ceval.c:

                WHY_NOT,        OK, not this one :-)
                WHY_EXCEPTION,  SApplicationException
                WHY_RERAISE,    implemented differently, see Reraise
                WHY_RETURN,     SReturnValue
                WHY_BREAK,      SBreakLoop
                WHY_CONTINUE,   SContinueLoop
                WHY_YIELD       not needed
    """
    _immutable_ = True
    def nomoreblocks(self):
        raise BytecodeCorruption("misplaced bytecode - should not return")

class SReturnValue(SuspendedUnroller):
    """Signals a 'return' statement.
    Argument is the wrapped object to return."""
    _immutable_ = True
    kind = 0x01
    def __init__(self, w_returnvalue):
        self.w_returnvalue = w_returnvalue
    def nomoreblocks(self):
        return self.w_returnvalue

class SApplicationException(SuspendedUnroller):
    """Signals an application-level exception
    (i.e. an OperationException)."""
    _immutable_ = True
    kind = 0x02
    def __init__(self, operr):
        self.operr = operr
    def nomoreblocks(self):
        raise RaiseWithExplicitTraceback(self.operr)

class SBreakLoop(SuspendedUnroller):
    """Signals a 'break' statement."""
    _immutable_ = True
    kind = 0x04
SBreakLoop.singleton = SBreakLoop()

class SContinueLoop(SuspendedUnroller):
    """Signals a 'continue' statement.
    Argument is the bytecode position of the beginning of the loop."""
    _immutable_ = True
    kind = 0x08
    def __init__(self, jump_to):
        self.jump_to = jump_to


class FrameBlock(object):
    """Abstract base class for frame blocks from the blockstack,
    used by the SETUP_XXX and POP_BLOCK opcodes."""

    _immutable_ = True

    def __init__(self, valuestackdepth, handlerposition, previous):
        self.handlerposition = handlerposition
        self.valuestackdepth = valuestackdepth
        self.previous = previous   # this makes a linked list of blocks

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                self.handlerposition == other.handlerposition and
                self.valuestackdepth == other.valuestackdepth)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.handlerposition, self.valuestackdepth))

    def cleanupstack(self, frame):
        frame.dropvaluesuntil(self.valuestackdepth)

    def pop_block(self, frame):
        "Clean up a frame when we normally exit the block."
        self.cleanupstack(frame)

    # internal pickling interface, not using the standard protocol
    def _get_state_(self, space):
        return space.newtuple([space.newtext(self._opname), space.newint(self.handlerposition),
                               space.newint(self.valuestackdepth)])

    def handle(self, frame, unroller):
        """ Purely abstract method
        """
        raise NotImplementedError

class LoopBlock(FrameBlock):
    """A loop block.  Stores the end-of-loop pointer in case of 'break'."""

    _immutable_ = True
    _opname = 'SETUP_LOOP'
    handling_mask = SBreakLoop.kind | SContinueLoop.kind

    def handle(self, frame, unroller):
        if isinstance(unroller, SContinueLoop):
            # re-push the loop block without cleaning up the value stack,
            # and jump to the beginning of the loop, stored in the
            # exception's argument
            frame.append_block(self)
            jumpto = unroller.jump_to
            ec = frame.space.getexecutioncontext()
            return r_uint(frame.jump_absolute(jumpto, ec))
        else:
            # jump to the end of the loop
            self.cleanupstack(frame)
            return r_uint(self.handlerposition)


class SysExcInfoRestorer(FrameBlock):
    """
    This is a special, implicit block type which is created when entering a
    finally or except handler. It does not belong to any opcode
    """

    _immutable_ = True
    _opname = 'SYS_EXC_INFO_RESTORER' # it's not associated to any opcode
    handling_mask = 0 # this block is never handled, only popped by POP_EXCEPT

    def __init__(self, operr, previous):
        self.operr = operr
        self.previous = previous

    def pop_block(self, frame):
        assert False # never called

    def handle(self, frame, unroller):
        assert False # never called

    def cleanupstack(self, frame):
        ec = frame.space.getexecutioncontext()
        ec.set_sys_exc_info(self.operr)


class ExceptBlock(FrameBlock):
    """An try:except: block.  Stores the position of the exception handler."""

    _immutable_ = True
    _opname = 'SETUP_EXCEPT'
    handling_mask = SApplicationException.kind

    def handle(self, frame, unroller):
        # push the exception to the value stack for inspection by the
        # exception handler (the code after the except:)
        self.cleanupstack(frame)
        # the stack setup is slightly different than in CPython:
        # instead of the traceback, we store the unroller object,
        # wrapped.
        assert isinstance(unroller, SApplicationException)
        operationerr = unroller.operr
        operationerr.normalize_exception(frame.space)
        frame.pushvalue(unroller)
        frame.pushvalue(operationerr.get_w_value(frame.space))
        frame.pushvalue(operationerr.w_type)
        # set the current value of sys_exc_info to operationerr,
        # saving the old value in a custom type of FrameBlock
        frame.save_and_change_sys_exc_info(operationerr)
        return r_uint(self.handlerposition)   # jump to the handler


class FinallyBlock(FrameBlock):
    """A try:finally: block.  Stores the position of the exception handler."""

    _immutable_ = True
    _opname = 'SETUP_FINALLY'
    handling_mask = -1     # handles every kind of SuspendedUnroller

    def handle(self, frame, unroller):
        # any abnormal reason for unrolling a finally: triggers the end of
        # the block unrolling and the entering the finally: handler.
        # see comments in cleanup().
        self.cleanupstack(frame)
        operationerr = None
        if isinstance(unroller, SApplicationException):
            operationerr = unroller.operr
            operationerr.normalize_exception(frame.space)
        frame.pushvalue(unroller)
        # set the current value of sys_exc_info to operationerr,
        # saving the old value in a custom type of FrameBlock
        frame.save_and_change_sys_exc_info(operationerr)
        return r_uint(self.handlerposition)   # jump to the handler

    def pop_block(self, frame):
        self.cleanupstack(frame)
        frame.save_and_change_sys_exc_info(None)


block_classes = {'SYS_EXC_INFO_RESTORER': SysExcInfoRestorer,
                 'SETUP_LOOP': LoopBlock,
                 'SETUP_EXCEPT': ExceptBlock,
                 'SETUP_FINALLY': FinallyBlock,
                 'SETUP_WITH': FinallyBlock,
                 }


##class W_OperationError(W_Root):
##    """
##    Tiny applevel wrapper around an OperationError.
##    """
##
##    def __init__(self, operr):
##        self.operr = operr
##
##    def descr_reduce(self, space):
##        from pypy.interpreter.mixedmodule import MixedModule
##        w_mod = space.getbuiltinmodule('_pickle_support')
##        mod = space.interp_w(MixedModule, w_mod)
##        w_new_inst = mod.get('operationerror_new')
##        w_args = space.newtuple([])
##        operr = self.operr
##        if operr is None:
##            return space.newtuple([w_new_inst, w_args])
##        w_state = space.newtuple([operr.w_type, operr.get_w_value(space),
##                                  operr.get_traceback()])
##        return space.newtuple([w_new_inst, w_args, w_state])
##
##    def descr_setstate(self, space, w_state):
##        w_type, w_value, w_tb = space.fixedview(w_state, 3)
##        self.operr = OperationError(w_type, w_value, w_tb)

def ensure_ns(space, w_globals, w_locals, funcname, caller=None):
    """Ensure globals/locals exist and are of the correct type"""
    if (not space.is_none(w_globals) and
        not space.isinstance_w(w_globals, space.w_dict)):
        raise oefmt(space.w_TypeError,
                    '%s() arg 2 must be a dict, not %T', funcname, w_globals)
    if (not space.is_none(w_locals) and
        space.lookup(w_locals, '__getitem__') is None):
        raise oefmt(space.w_TypeError,
                    '%s() arg 3 must be a mapping or None, not %T',
                    funcname, w_locals)

    if space.is_none(w_globals):
        if caller is None:
            caller = space.getexecutioncontext().gettopframe_nohidden()
        if caller is None:
            w_globals = space.newdict(module=True)
            if space.is_none(w_locals):
                w_locals = w_globals
        else:
            w_globals = caller.get_w_globals()
            if space.is_none(w_locals):
                w_locals = caller.getdictscope()
    elif space.is_none(w_locals):
        w_locals = w_globals

    return w_globals, w_locals



__docformat__ = "restructuredtext"

from xpython.vm import (
    PyVM,
    PyVMError,
    PyVMRuntimeError,
    )

from xpython.vmtrace import (
    PyVMTraced,
    )

from xpython.pyobj import (
    Function,
    Method,
    Cell,
    Traceback,
    traceback_from_frame,
    Generator
    )

from xpython.byteop import (
    get_byteop,
    )
